'''
    Copyright 2018 SerialLab, CORP

    This file is part of ListBasedFlavorpack.

    RandomFlavorpack is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RandomFlavorpack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with RandomFlavorpack.  If not, see <http://www.gnu.org/licenses/>.
'''
from django.contrib import admin
from list_based_flavorpack import models


class ProcessingParametersInline(admin.TabularInline):
    model = models.ProcessingParameters
    extra = 0


class ListBasedRegionAdmin(admin.ModelAdmin):
    list_display = (
        'readable_name',
        'machine_name',
        'pool',
        'file_id'
    )
    inlines = [
        ProcessingParametersInline
    ]
    ordering = [
        'readable_name'
    ]


def register_to_site(admin_site):
    admin_site.register(models.ListBasedRegion, ListBasedRegionAdmin)



import os
import linecache
from django.test import TestCase
from django.core.management import call_command
from django.test.client import RequestFactory
from quartet_templates.models import Template
from serialbox.discovery import get_generator
from serialbox.api import serializers as sb_serializers
from list_based_flavorpack.models import ListBasedRegion, ProcessingParameters
from serialbox import models as sb_models
from quartet_output.models import EndPoint, AuthenticationInfo
from quartet_capture.models import Rule, Step
from quartet_templates.models import Template


class TemplateTest(TestCase):
    '''
    Tests the Template model.
    '''
    def create_template(self, name: str, content: str, description: str):
        return Template.objects.create(name=name, content=content, description=description)

    def test_template_creation(self):
        content = '''
<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns:urn="urn:bla:bla">
   <soapenv:Header/>
   <soapenv:Body>
      <urn:SpecialXML>
         <SomethingSpecial>{{ something_special }}</SomethingSpecial>
      </urn:SpecialXML>
   </soapenv:Body>
</soapenv:Envelope>
        '''
        template = self.create_template(name="Test Template", content=content, description="A Test Template")
        self.assertTrue(isinstance(template, Template))

    def test_template_rendering(self):
        content = '''
<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns:urn="urn:bla:bla">
   <soapenv:Header/>
   <soapenv:Body>
      <urn:SpecialXML>
         <SomethingSpecial>{{ something_special }}</SomethingSpecial>
      </urn:SpecialXML>
   </soapenv:Body>
</soapenv:Envelope>
        '''
        template = self.create_template(name="Test Template", content=content, description="A Test Template")
        rendered = template.render(context={"something_special": "Special Value"})
        expected = '''
<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns:urn="urn:bla:bla">
   <soapenv:Header/>
   <soapenv:Body>
      <urn:SpecialXML>
         <SomethingSpecial>Special Value</SomethingSpecial>
      </urn:SpecialXML>
   </soapenv:Body>
</soapenv:Envelope>
        '''
        self.assertEqual(expected, rendered)


class ListBasedRegionTest(TestCase):
    '''
    Tests the creation of a list-based region.
    '''
    
    def test_list_based_region_creation(self):
        content = '''
<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns:urn="urn:bla:bla">
   <soapenv:Header/>
   <soapenv:Body>
      <urn:SpecialXML>
         <RequestSize>{{ allocate.size }}</RequestSize>
      </urn:SpecialXML>
   </soapenv:Body>
</soapenv:Envelope>
        '''
        template =  Template.objects.create(name="test template", content=content, description="A test template")
        list_based_region = ListBasedRegion()
        list_based_region.template = template
        list_based_region.processing_class_path = "list_based_flavorpack.processing_classes.third_party_processing.processing.ThirdPartyProcessingClass"
        list_based_region.directory_path = "/tmp"
        list_based_region.number_replenishment_size = 20
        list_based_region.save()
        self.assertEqual("/tmp/" + str(list_based_region.file_id), list_based_region.file_path)


class UUIDPoolTest(TestCase):
    '''
    Tests the Third Party Processing Class along with the UUID generation step.
    '''
    def setUp(self):
        self.test_pool = self.generate_test_pool()
        self.rule = self.generate_step_rule()
        self.template = self.generate_template()
        self.list_based_region = self.generate_region(self.test_pool, self.rule, self.template)

    def tearDown(self):
        try:
            os.remove(self.list_based_region.file_path)
        except:
            # some tests don't use filesystem.
            pass
    
    def generate_test_pool(self):
        # create pool
        test_pool = sb_models.Pool()
        test_pool.readable_name = "Test Pool"
        test_pool.machine_name = "TestPool"
        test_pool.active = True
        test_pool.request_threshold = 1000
        test_pool.save()
        return test_pool
    
    def generate_region(self, test_pool, rule, template):
        # create region with third party processing class.
        list_based_region = ListBasedRegion()
        list_based_region.processing_class_path = "list_based_flavorpack.processing_classes.third_party_processing.processing.ThirdPartyProcessingClass"
        list_based_region.pool = test_pool
        list_based_region.readable_name = "UUID Region"
        list_based_region.machine_name = "UUIDRegion"
        list_based_region.active = True
        list_based_region.order = 1
        list_based_region.rule = rule
        list_based_region.template = template
        list_based_region.number_replenishment_size = 200
        list_based_region.directory_path = "/tmp"
        list_based_region.save()
        return list_based_region
    
    def generate_template(self):
        content = '''<?xml version="1.0" encoding="UTF-8"?>
<NumberRequest>
    <Type>UUID</Type>
    <Size>{{ allocate.size }}</Size>
</NumberRequest>
        '''
        # create template
        template = Template.objects.create(name="UUID Request", content=content, description="Requests a UUID")
        return template
    
    def generate_step_rule(self):
        # Create Rule and Step
        rule = Rule()
        rule.name = "UUID Rule"
        
        rule.description = "Generates UUIDs to a file"
        rule.save()
        step = Step()
        step.name = "UUID Generate Step"
        step.description = "Generates UUIDs to a file"
        step.step_class = "list_based_flavorpack.steps.UUIDRequestStep"
        step.order = 1
        step.rule = rule
        step.save()
        return rule
    
    def generate_allocation(self, size, test_pool, list_based_region):
        generator = get_generator(test_pool.machine_name)
        request_factory = RequestFactory()
        request = request_factory.get("allocate/TestPool/" + str(size))
        response = generator.get_response(request, size, test_pool.machine_name, None)
        serializer = sb_serializers.ResponseSerializer(response)
        return response

    def test_uuid_pool_correct_filepath(self):
        self.assertEqual("/tmp/" + str(self.list_based_region.file_id), self.list_based_region.file_path)
        
    def test_uuid_pool_200_requested(self):
        self.assertEqual("/tmp/" + str(self.list_based_region.file_id), self.list_based_region.file_path)
        # Now request 5 UUIDs, the file should be written with 200 UUIDs
        size = 5
        response = self.generate_allocation(size, self.test_pool, self.list_based_region)
        # check that there are 200 items written in the file (number replenishment size)
        file_size = sum((1 for i in open(self.list_based_region.file_path)))
        self.assertEqual(200, file_size)
   
    def test_uuid_pool_correct_numbers_returned(self):
        response = self.generate_allocation(5, self.test_pool, self.list_based_region)
        numbers_returned = response.number_list
        # check first five items requested.
        with open(self.list_based_region.file_path, 'r') as f:
            count = 0
            for line in f.readlines():
                self.assertEqual(numbers_returned[count], line.strip())
                count += 1
                if count == 5:
                    break

    def test_uuid_pool_correct_numbers_returned_2_calls(self):
        linecache.checkcache(filename=self.list_based_region.file_path)
        response = self.generate_allocation(5, self.test_pool, self.list_based_region)
        numbers_returned = response.number_list
        # check first five items requested.
        max_line = self.list_based_region.last_number_line + 5
        count = 0
        count = 0
        for lineno in range(1, 5):
            self.assertEqual(numbers_returned[count], linecache.getline(self.list_based_region.file_path, lineno).strip())
            count += 1
        response = self.generate_allocation(10, self.test_pool, self.list_based_region)
        numbers_returned = response.number_list
        count = 0
        for lineno in range(6, 14):
            self.assertEqual(numbers_returned[count], linecache.getline(self.list_based_region.file_path, lineno).strip())
            count += 1

    def test_uuid_pool_correct_numbers_returned_5_calls_replenish_once(self):
        all_numbers_returned = []
        linecache.checkcache(filename=self.list_based_region.file_path)
        response = self.generate_allocation(5, self.test_pool, self.list_based_region)
        # with a request of 5, we should have 200 items in the file.
        file_size = sum((1 for i in open(self.list_based_region.file_path)))
        self.assertEqual(200, file_size)
        numbers_returned = response.number_list
        all_numbers_returned += response.number_list
        # check first five items requested.
        count = 0
        for lineno in range(1, 6):
            self.assertEqual(numbers_returned[count], linecache.getline(self.list_based_region.file_path, lineno).strip())
            count += 1
        response = self.generate_allocation(10, self.test_pool, self.list_based_region)

        # with another request of 10, we should still have 200 in the file.
        file_size = sum((1 for i in open(self.list_based_region.file_path)))
        self.assertEqual(200, file_size)        
        numbers_returned = response.number_list
        all_numbers_returned += response.number_list
        count = 0
        linecache.checkcache(filename=self.list_based_region.file_path)        
        for lineno in range(6, 16):
            self.assertEqual(numbers_returned[count], linecache.getline(self.list_based_region.file_path, lineno).strip())
            count += 1
        response = self.generate_allocation(100, self.test_pool, self.list_based_region)
        # with another request of 100, we should still just have 200 in the file.        
        file_size = sum((1 for i in open(self.list_based_region.file_path)))
        self.assertEqual(200, file_size)        
        numbers_returned = response.number_list
        all_numbers_returned += response.number_list
        count = 0
        linecache.checkcache(filename=self.list_based_region.file_path)        
        for lineno in range(16, 101):
            self.assertEqual(numbers_returned[count], linecache.getline(self.list_based_region.file_path, lineno).strip())
            count += 1
        # replenish now. Next file size should be 400.
        response = self.generate_allocation(150, self.test_pool, self.list_based_region)
        numbers_returned = response.number_list
        all_numbers_returned += response.number_list
        count = 0
        '''
        print(numbers_returned)
        with open(self.list_based_region.file_path, 'r') as f:
            for line in f.readlines():
                count += 1
                print("count: %s" % str(count))
                print(line)
        '''
        count = 0
        # Replenishment should have been triggered and the file should have 400 UUIDs.
        file_size = sum((1 for i in open(self.list_based_region.file_path)))
        self.assertEqual(400, file_size)
        linecache.checkcache(filename=self.list_based_region.file_path)        
        for lineno in range(116, 266):
            self.assertEqual(numbers_returned[count], linecache.getline(self.list_based_region.file_path, lineno).strip())
            count += 1

        response = self.generate_allocation(50, self.test_pool, self.list_based_region)
        numbers_returned = response.number_list
        all_numbers_returned += response.number_list
        count = 0
        '''
        print(numbers_returned)
        with open(self.list_based_region.file_path, 'r') as f:
            for line in f.readlines():
                count += 1
                print("count: %s" % str(count))
                print(line)
        '''
        count = 0
        # with another request of 50, we should still have 400 UUIDs in the file.
        file_size = sum((1 for i in open(self.list_based_region.file_path)))
        self.assertEqual(400, file_size)
        linecache.checkcache(filename=self.list_based_region.file_path)        
        for lineno in range(266, 316):
            self.assertEqual(numbers_returned[count], linecache.getline(self.list_based_region.file_path, lineno).strip())
            count += 1

        # Finally overall check the continuity of numbers returned vs numbers in the file.
        # make a list of the 315 items in the file.
        file_numbers_returned = []
        for lineno in range(1, 316):
            file_numbers_returned.append(linecache.getline(self.list_based_region.file_path, lineno).strip())
        self.assertEqual(len(all_numbers_returned), len(file_numbers_returned))
        for i in range(len(file_numbers_returned)):
            self.assertEqual(all_numbers_returned[i], file_numbers_returned[i])



# Generated by Django 2.1.1 on 2018-09-27 17:16

import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('serialbox', '0001_initial'),
        ('quartet_output', '0002_auto_20180909_0946'),
        ('quartet_capture', '0001_initial'),
        ('quartet_templates', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ListBasedRegion',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_date', models.DateTimeField(auto_now_add=True, db_index=True, help_text='The date and time that this record was created', verbose_name='Created Date')),
                ('modified_date', models.DateTimeField(auto_now=True, db_index=True, help_text='The date and time that this record was modified last.', verbose_name='Last Modified')),
                ('readable_name', models.CharField(help_text='A human-readable name for use in GUIs and reports and such.', max_length=100, unique=True, verbose_name='Readable Name')),
                ('machine_name', models.CharField(help_text='A url/api-friendly unique key for use in API calls and such.', max_length=100, unique=True, validators=[django.core.validators.RegexValidator('^[A-Za-z0-9]*$', 'Only numbers and letters are allowed. Invalid API Key.')], verbose_name='API Key')),
                ('active', models.BooleanField(default=True, help_text='Whether or not this pool is active/in-use. If marked false the pool will no longer be able to be used in API calls, etc.', verbose_name='Active')),
                ('order', models.IntegerField(blank=True, help_text='The order in which this region will be consumed as numbers are issued from the pool overall', null=True, verbose_name='Order')),
                ('last_number_line', models.IntegerField(default=1, help_text='The line number of the last number issued maintains the state of the list-based region.', verbose_name='Last Number Line')),
                ('processing_class_path', models.CharField(help_text='The full python path to the class that will be processing region allocations', max_length=150, verbose_name='Processing Class Path')),
                ('file_id', models.UUIDField(default=uuid.uuid1, editable=False)),
                ('directory_path', models.CharField(blank=True, default='/var/quartet/numbers', help_text='The full path to the directory where numbers will be stored. Leave blank for default. Default is /var/quartet/numbers', max_length=150, null=True)),
                ('number_replenishment_size', models.IntegerField(help_text='The size that the outbound message will request from the third-party system, if numbers available are low. E.g.: 500', verbose_name='Number Replenishment Size')),
                ('authentication_info', models.ForeignKey(help_text='The Authentication Info to use.', null=True, on_delete=django.db.models.deletion.PROTECT, to='quartet_output.AuthenticationInfo', verbose_name='Authentication Info')),
                ('end_point', models.ForeignKey(help_text='A protocol-specific endpoint defining whereany data will come from', null=True, on_delete=django.db.models.deletion.PROTECT, to='quartet_output.EndPoint', verbose_name='End Point')),
                ('pool', models.ForeignKey(blank=True, help_text='The Number Pool this region will belong to.', null=True, on_delete=django.db.models.deletion.CASCADE, to='serialbox.Pool', verbose_name='Number Pool')),
                ('rule', models.ForeignKey(help_text='A rule that may be executed by the region processing class.', null=True, on_delete=django.db.models.deletion.PROTECT, to='quartet_capture.Rule', verbose_name='Processing Rule')),
                ('template', models.ForeignKey(help_text='The Django/Jinja template to send a formatted request for number ranges', null=True, on_delete=django.db.models.deletion.PROTECT, to='quartet_templates.Template', verbose_name='Message Template')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProcessingParameters',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(help_text='The key part of the key-value pair. Example: q', max_length=200)),
                ('value', models.CharField(blank=True, help_text='The value part of the key-value pair.', max_length=400, null=True)),
                ('list_based_region', models.ForeignKey(help_text='A key-value pair object meant to hold parameters used for processing classes.', null=True, on_delete=django.db.models.deletion.PROTECT, related_name='processing_parameters', to='list_based_flavorpack.ListBasedRegion', verbose_name='Processing Class Parameter')),
            ],
        ),
        migrations.AlterUniqueTogether(
            name='listbasedregion',
            unique_together={('pool', 'order')},
        ),
    ]



'''
    Copyright 2018 SerialLab, CORP

    This file is part of ListBasedFlavorpack.

    RandomFlavorpack is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    RandomFlavorpack is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with RandomFlavorpack.  If not, see <http://www.gnu.org/licenses/>.
'''
import os
import uuid
from serialbox.rules.common import PreprocessingRule
from serialbox.rules.errors import RuleError
from quartet_capture.tasks import create_and_queue_task
from quartet_capture.models import TaskParameter


class ValidDirectoryError(RuleError):

    def __init__(self, detail=None, directory_path=""):
        self.default_detail = ('An error occurred when attempting to find/create '
                               'the directory to store numbers: %s' % directory_path)
        RuleError.__init__(self, detail=detail)


class ValidNumberDirectory(PreprocessingRule):
    '''
    Checks if the number directory exists and attempts to create it if applicable.
    '''
    def execute(self, request, pool, region, size):
        # check directory_path
        try:
            if not os.path.exists(region.directory_path):
                os.makedirs(region.directory_path)
            if not os.path.exists(region.file_path):
                with open(region.file_path, 'a'):
                    os.utime(region.file_path, None)
        except:
            raise ValidDirectoryError(directory_path=region.file_path)


class SufficientNumbersStorage(PreprocessingRule):
    '''
    This rule will check whether there is a sufficient amount of numbers for the size given,
    and if not, will make a request to the third party system to receive enough numbers.
    If an insufficient number of numbers or no number is returned, it will throw an error.
    '''

    def fetch_more_numbers(self, request, pool, region, size):
        '''
        Gets more numbers from the third-party system and writes them to the
        local file store.
        '''
        rule_name = region.rule.name
        template = region.template
        task_param = TaskParameter(name="List-based Region", value=region.machine_name)
        processing_params_dict = {item.key:item.value for item in region.processing_parameters.all()}

        processing_params_dict["authentication_info"] = region.authentication_info
        
        if region.number_replenishment_size > size:
            processing_params_dict["allocate"] = {"size": region.number_replenishment_size,
                                                  "random_event_id": str(uuid.uuid1())}
        else:
            processing_params_dict["allocate"] = {"size": size, "random_event_id": str(uuid.uuid1())}

        # template rendered
        rendered = template.render(processing_params_dict)
        task = create_and_queue_task(rendered, rule_name, task_type = "Input", run_immediately=True, task_parameters=[task_param])
        task.refresh_from_db()  # ensures we get the processed result for the task.a
        region.refresh_from_db()  # ensures we have the most udpated version (in case of parallel calls)
        if task.status != "FINISHED":
            raise RuleError(detail="An error occurred while attempting to request new numbers from third-party system. Please check the log output from task %s" % task.name)
    
    def execute(self, request, pool, region, size):
        try:
            file_size = sum((1 for i in open(region.file_path)))
        except OSError as e:
            raise RuleError(detail="An error occurred while opening attempting to "
                            "read the file to store numbers: %s" % region.file_path)
        if region.last_number_line + size >= file_size:
            currently_available = (file_size + 1) - region.last_number_line  # don't overfetch if not needed.
            self.fetch_more_numbers(request, pool, region, size - currently_available)
        else:
            # numbers are sufficient. Nothing to do.
            pass



# This program is free software: you can redistribute it and/| modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, |
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY | FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Copyright 2019 SerialLab Corp.  All rights reserved.
import calendar
import time
from serialbox import models as sb_models
from list_based_flavorpack import models as lb_models
from copy import deepcopy
from uuid import uuid1


def clone_list_based_pool(pool_machine_name, new_pool_machine_name):
    """
    Clones a list-based pool.
    :param pool_machine_name:  The machine name of the pool to clone.
    :param new_pool_machine_name: A new machine name for the new, cloned, pool.
    :return:
    """
    pool = sb_models.Pool.objects.get(machine_name=pool_machine_name)
    region = lb_models.ListBasedRegion.objects.get(pool=pool)
    new_pool = deepcopy(pool)
    new_region = deepcopy(region)
    new_pool.id = None
    new_pool.readable_name = new_pool.readable_name + ' clone_' + str(calendar.timegm(time.gmtime()))
    new_pool.machine_name = new_pool_machine_name
    new_pool.save()
    response_rules = sb_models.ResponseRule.objects.filter(pool=pool)
    for rrule in response_rules:
        new_rrule = deepcopy(rrule)
        new_rrule.id = None
        new_rrule.pool = new_pool
        new_rrule.save()
    new_region.id = None
    new_region.machine_name = new_pool_machine_name
    new_region.readable_name = new_region.readable_name + ' clone '
    new_region.active = False
    new_region.pool = new_pool
    new_region.file_id = uuid1()
    new_region.last_line_number = 1
    new_region.save()
    for param in region.processing_parameters.all():
        new_param = deepcopy(param)
        new_param.id = None
        new_param.list_based_region = new_region
        if param.value == pool_machine_name:
            new_param.value = new_pool_machine_name
        new_param.save()


from setuptools import setup, find_packages

setup(
    name='list_based_flavorpack',
    version='2.0.3',
    packages=find_packages(),
    include_package_data=True,
    url='https://gitlab.com/serial-lab/list_based_flavorpack',
    license='GPLv3',
    author='SerialLab Corp',
    author_email='slab@serial-lab.com',
    description='Generic List-Based Flavorpack Plugin for SerialBox',
    keywords=('seriallab, quartet, list-based, list, serialbox, '
             'level-4 quartet'),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



import linecache
from list_based_flavorpack.processing_classes.third_party_processing import rules

class ThirdPartyProcessingClass:
    '''
    An example of a processing  class for a list based pool/region.
    Returns UUIDs stored in memory.
    '''
    def __init__(self):
        '''
        Sets processing rules.
        '''
        self.pre_processing_rules = [rules.ValidNumberDirectory, rules.SufficientNumbersStorage]
        self.post_processing_rules = []
    
    def get_pre_processing_rules(self):
        '''
        Returns processing rules to be executed prior to the generator running.
        '''
        return self.pre_processing_rules
    
    def get_post_processing_rules(self):
        '''
        Returns processing rules to be executed after the generator has run.
        '''
        return self.post_processing_rules

    def get_items(self, region, size):
        '''
        Pull numbers from the file created/updated from rules previously executed.
        '''
        first = region.last_number_line
        max_line = first + size
        lines = []
        linecache.checkcache(filename=region.file_path)
        for lineno in range(first, max_line):
            item = linecache.getline(region.file_path, lineno).strip()
            if not item:
                raise ValueError("There is not enough number available for this iteration on line number: %s", lineno)
            lines.append(item)
        region.last_number_line = max_line
        region.save()
        return lines
